/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.binding.soap;

import java.io.File;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.QName;

import org.objectweb.petals.common.descriptor.JBIDescriptor;

/**
 * The SOAP component context. The context is filled by the SU listener (adding
 * modules, service descriptions...) and used by the listeners.
 * 
 * @author Christophe HAMERLING - eBMWebSourcing
 * 
 */
public class SoapComponentContext {

    /**
     * The map of JBI descriptors. Key is the service unit name.
     */
    private Map<String, JBIDescriptor> jbiDescriptors;

    /**
     * The map of services.xml files. Key is the service unit name.
     */
    private Map<String, File> servicesDescriptors;

    /**
     * The requested modules to engage for the service client. Key is the
     * address, List is the QNames of the modules.
     */
    private Map<String, List<QName>> modules;
    
    /**
     * Creates a new instance of {@link SoapComponentContext}
     * 
     */
    public SoapComponentContext() {
        this.jbiDescriptors = new HashMap<String, JBIDescriptor>();
        this.servicesDescriptors = new HashMap<String, File>();
        this.modules = new HashMap<String, List<QName>>();
    }

    /**
     * Add a {@link JBIDescriptor}
     * 
     * @param address
     * @param jbiDescriptor
     */
    public void addJbiDescriptor(String serviceUnitName,
        JBIDescriptor jbiDescriptor) {
        if (serviceUnitName == null) {
            throw new IllegalArgumentException(
                "Service unit name could not be null");
        }
        jbiDescriptors.put(serviceUnitName, jbiDescriptor);
    }

    /**
     * Get the JBI descriptor for the given address.
     * 
     * @param address
     * @return the {@link JBIDescriptor} if found, else return null
     */
    public JBIDescriptor getJbiDescriptor(String serviceUnitName) {
        return jbiDescriptors.get(serviceUnitName);
    }

    /**
     * Remove the {@link JBIDescriptor} for the given address
     * 
     * @param address
     */
    public void removeJbiDescriptor(String serviceUnitName) {
        if (jbiDescriptors != null) {
            jbiDescriptors.remove(serviceUnitName);
        }
    }

    /**
     * Add a {@link JBIDescriptor}
     * 
     * @param address
     * @param jbiDescriptor
     */
    public void addServiceDescriptor(String serviceUnitName,
        File serviceDescriptor) {
        if (serviceUnitName == null) {
            throw new IllegalArgumentException(
                "Service unit name could not be null");
        }
        servicesDescriptors.put(serviceUnitName, serviceDescriptor);
    }

    /**
     * Get the service descriptor as {@link File} of the given service unit if
     * available.
     * 
     * @param address
     * @return the file (services.xml) or null if no service descriptor is
     *         available
     */
    public File getServiceDescriptor(String serviceUnitName) {
        return servicesDescriptors.get(serviceUnitName);
    }

    /**
     * Remove the {@link JBIDescriptor} for the given service unit
     * 
     * @param address
     */
    public void removeServiceDescriptor(String serviceUnitName) {
        if (servicesDescriptors != null) {
            servicesDescriptors.remove(serviceUnitName);
        }
    }

    /**
     * Add a {@link JBIDescriptor}
     * 
     * @param address
     * @param jbiDescriptor
     */
    public void addModules(String address, List<QName> list) {
        if (address == null) {
            throw new IllegalArgumentException("Address could not be null");
        }
        modules.put(address, list);
    }

    /**
     * Get the modules for the given address.
     * 
     * @param address
     * @return the modules of the given address. Should at least contains the
     *         addressing module
     */
    public List<QName> getModules(String address) {
        return modules.get(address);
    }

    /**
     * Remove the {@link JBIDescriptor} for the given address
     * 
     * @param address
     */
    public void removeModules(String address) {
        if (modules != null) {
            modules.remove(address);
        }
    }
}
