/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.binding.soap;

import java.io.File;
import java.io.FileInputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.component.ComponentContext;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamException;

import org.apache.axis2.AxisFault;
import org.apache.axis2.Constants;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.deployment.ServiceBuilder;
import org.apache.axis2.description.AxisService;
import org.apache.axis2.description.Parameter;
import org.apache.axis2.engine.AxisConfiguration;
import org.objectweb.petals.binding.soap.listener.SUProperties;
import org.objectweb.petals.binding.soap.listener.incoming.PetalsAxisService;
import org.objectweb.petals.common.descriptor.Consumes;
import org.objectweb.petals.common.descriptor.Extensions;
import org.objectweb.petals.common.descriptor.JBIDescriptor;
import org.objectweb.petals.common.descriptor.Provides;
import org.objectweb.petals.component.common.PEtALSComponentSDKException;
import org.objectweb.petals.component.common.su.ServiceUnitListener;
import org.objectweb.petals.component.common.util.PetalsExtensionsUtil;

/**
 * A service unit listener used to register new service into Axis Engine during
 * SU deployment.
 *
 * @author Christophe HAMERLING - eBMWebSourcing
 *
 */
public class SoapSUListener implements ServiceUnitListener {

    private ConfigurationContext axisConfigurationContext;

    private ComponentContext componentContext;

    private SoapComponentContext soapContext;

    private Logger logger;

    private String servicesPrefix;

    /**
     * Creates a new instance of {@link SoapSUListener}
     *
     * @param axisConfigurationContext
     * @param componentContext
     * @param logger
     */
    public SoapSUListener(SoapComponentContext soapContext,
            ConfigurationContext axisConfigurationContext,
            ComponentContext componentContext, String servicesURL, Logger logger) {
        this.soapContext = soapContext;
        this.axisConfigurationContext = axisConfigurationContext;
        this.componentContext = componentContext;
        this.servicesPrefix = servicesURL;
        this.logger = logger;
    }

    /*
     * (non-Javadoc)
     *
     * @see org.objectweb.petals.component.common.su.ServiceUnitListener#onSUDeployed(java.lang.String,
     *      java.lang.String,
     *      org.objectweb.petals.tools.jbicommon.descriptor.JBIDescriptor)
     */
    public void onSUDeployed(String serviceUnitName, String suRootPath,
            JBIDescriptor descriptor) throws PEtALSComponentSDKException {
        logger.log(Level.FINE, "Deploying a new AxisService for SU '"
                + serviceUnitName + "'");

        soapContext.addJbiDescriptor(serviceUnitName, descriptor);

        // check if a services.xml file is available
        File servicesXml = new File(suRootPath, "services.xml");
        if (servicesXml.exists()) {
            soapContext.addServiceDescriptor(serviceUnitName, servicesXml);
        }

        // add the modules references to the context
        addModulesToContext(descriptor);
    }

    /*
     * (non-Javadoc)
     *
     * @see org.objectweb.petals.component.common.su.ServiceUnitListener#onSUStarted(java.lang.String)
     */
    public void onSUStarted(String serviceUnitName)
            throws PEtALSComponentSDKException {

        // get descriptors
        JBIDescriptor descriptor = soapContext
                .getJbiDescriptor(serviceUnitName);
        File serviceDescription = soapContext
                .getServiceDescriptor(serviceUnitName);

        if (descriptor != null) {
            // register new Axis2 service only on consumes nodes
            List<Consumes> consumesList = descriptor.getServices()
                    .getConsumes();
            /*
             * Create an AxisService for each requested endpoint. If an enpoint
             * has not been found, simply log a severe error. TODO : If a
             * service creation fails, and if services have been already
             * registered for this service unit, unregister them.
             */
            for (Consumes consumes : consumesList) {
                createAxisService(consumes, serviceDescription);
            }
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.objectweb.petals.component.common.su.ServiceUnitListener#onSUStopped(java.lang.String)
     */
    public void onSUStopped(String serviceUnitName)
            throws PEtALSComponentSDKException {

        JBIDescriptor descriptor = soapContext
                .getJbiDescriptor(serviceUnitName);
        if (descriptor != null) {
            // delete registered axis services
            List<Consumes> consumesList = descriptor.getServices()
                    .getConsumes();
            for (Consumes consumes : consumesList) {
                unregisterAxisService(consumes);
            }
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.objectweb.petals.component.common.su.ServiceUnitListener#onSUUndeployed(java.lang.String,
     *      java.lang.String,
     *      org.objectweb.petals.tools.jbicommon.descriptor.JBIDescriptor)
     */
    public void onSUUndeployed(String serviceUnitName, String suRootPath,
            JBIDescriptor descriptor) throws PEtALSComponentSDKException {

        soapContext.removeJbiDescriptor(serviceUnitName);
        soapContext.removeServiceDescriptor(serviceUnitName);

        removeModulesFromContext(descriptor);
    }

    /**
     * Create the Axis service that will handle incoming SOAP calls.
     *
     * @param consumes
     * @param serviceDescriptor
     * @throws PEtALSComponentSDKException
     */
    private void createAxisService(Consumes consumes, File serviceDescriptor)
            throws PEtALSComponentSDKException {

        QName serviceQName = consumes.getServiceName();
        String newServiceName = PetalsExtensionsUtil
                .extractValueFromKeyValueExtension(consumes.getExtensions(),
                        SUProperties.ADDRESS);

        logger.log(Level.INFO, "Registering '" + newServiceName
                + "' Axis service");

        AxisConfiguration axisConfig = axisConfigurationContext
                .getAxisConfiguration();

        try {
            AxisService axisService = null;
            String consumedEndpoint = consumes.getEndpointName();

            // try to find if the axisService has already been registered
            axisService = axisConfig.getService(newServiceName);

            if (axisService == null) {
                /*
                 * Create a new axisService, according to the JBI
                 * servicedescription. The endpoint name is the value given in
                 * the address extension, the namespace is provided in
                 * serviceQName and the endpoint is the one that will be really
                 * consumed.
                 */
                ServiceEndpoint endpoint = componentContext.getEndpoint(
                        serviceQName, consumedEndpoint);

                if (endpoint != null) {

                    axisService = new PetalsAxisService(newServiceName,
                            componentContext, consumes, servicesPrefix);

                    /*
                     * Add the modules for this new service. If the services.xml
                     * file if available, the modules references can be
                     * redefined
                     */
                    List<QName> modules = soapContext
                            .getModules(newServiceName);
                    if (modules != null) {
                        for (QName module : modules) {
                            axisService.addModuleref(module);
                        }
                    }

                    // Try to loas axis configuration from services.xml
                    if (serviceDescriptor != null) {
                        try {
                            FileInputStream fis = new FileInputStream(
                                    serviceDescriptor);
                            ServiceBuilder builder = new ServiceBuilder(fis,
                                    axisConfigurationContext, axisService);
                            builder.populateService(builder.buildOM());
                        } catch (XMLStreamException e) {
                            logger.warning(e.getMessage());
                        } catch (Exception e) {
                            logger.warning(e.getMessage());
                        }
                    }

                    // populate service with service descriptor
                    axisService.setTargetNamespace(serviceQName
                            .getNamespaceURI());
                    axisService.setEndpoint(consumedEndpoint);

                    axisService.addParameter(new Parameter(
                            Constants.SERVICE_CLASS, "PetalsReceiver"));

                    // add service
                    axisConfig.addService(axisService);

                } else {
                    logger.log(Level.SEVERE, "Endpoint '" + consumedEndpoint
                            + "' has not been found. The AxisService '"
                            + newServiceName + "' can not be activated");
                }
            } else {
                logger
                        .log(
                                Level.WARNING,
                                "The service '"
                                        + axisService.getName()
                                        + "' is already registered in Axis, you can not register it twice");
            }
        } catch (AxisFault e) {
            throw new PEtALSComponentSDKException(
                    "Can not register Service into Axis context", e);
        }
    }

    /**
     * Unregister the service from Axis
     *
     * @param serviceQName
     * @throws PEtALSComponentSDKException
     */
    private void unregisterAxisService(Consumes consumes)
            throws PEtALSComponentSDKException {

        // QName serviceQName = consumes.getServiceName();
        String endPointName = PetalsExtensionsUtil
                .extractValueFromKeyValueExtension(consumes.getExtensions(),
                        SUProperties.ADDRESS);

        AxisConfiguration axisConfig = axisConfigurationContext
                .getAxisConfiguration();

        logger.log(Level.INFO, "Removing Axis service '" + endPointName + "'");

        try {
            // register an axis service to axis engine
            AxisService axisService = axisConfig.getService(endPointName);

            // Feature request : #306664
            // FIXME : We do not have to remove the service group. It is
            // temporary until the next version of Axis2 where the removeService
            // method will fix it.
            if (axisService != null) {
                axisConfig.removeServiceGroup(endPointName);
                axisService.getAxisConfiguration().removeService(endPointName);
            } else {
                logger.log(Level.WARNING, "Service '" + endPointName
                        + "' not found, can not be unregistered from Axis2");
            }

        } catch (AxisFault e) {
            throw new PEtALSComponentSDKException(
                    "Can not remove service from Axis context", e);
        }
    }

    /**
     * Add the modules to the soap context
     *
     * @param serviceUnitName
     * @param descriptor
     */
    private void addModulesToContext(JBIDescriptor descriptor) {
        List<Provides> providesList = descriptor.getServices().getProvides();
        for (Provides provides : providesList) {
            addModules(provides.getExtensions());
        }

        List<Consumes> consumesList = descriptor.getServices().getConsumes();
        for (Consumes consumes : consumesList) {
            addModules(consumes.getExtensions());
        }
    }

    /**
     * Remove the modules from the soap context
     *
     * @param descriptor
     */
    private void removeModulesFromContext(JBIDescriptor descriptor) {
        List<Provides> providesList = descriptor.getServices().getProvides();
        for (Provides provides : providesList) {
            removeModules(provides.getExtensions());
        }
        List<Consumes> consumesList = descriptor.getServices().getConsumes();
        for (Consumes consumes : consumesList) {
            removeModules(consumes.getExtensions());
        }
    }

    /**
     * Add modules to the context
     *
     * @param provides
     */
    private void addModules(Extensions extensions) {
        String address = PetalsExtensionsUtil
                .extractValueFromKeyValueExtension(extensions,
                        SUProperties.ADDRESS);

        // get modules from extension
        String token = PetalsExtensionsUtil.extractValueFromKeyValueExtension(
                extensions, "modules");

        if (token != null) {
            // get individual modules values
            StringTokenizer st = new StringTokenizer(token, ",");
            List<QName> mods = new ArrayList<QName>(st.countTokens() + 1);
            mods.add(new QName(Constants.MODULE_ADDRESSING));
            while (st.hasMoreTokens()) {
                mods.add(new QName(st.nextToken().trim()));
            }
            soapContext.addModules(address, mods);
        }
    }

    /**
     * Remove the modules from the context.
     *
     * @param provides
     */
    private void removeModules(Extensions extensions) {
        String address = PetalsExtensionsUtil
                .extractValueFromKeyValueExtension(extensions,
                        SUProperties.ADDRESS);

        soapContext.removeModules(address);
    }
}
