/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2006 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: Axis2BCListener.java 154 19 avr. 2006 wjoseph $
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.binding.soap.listener.incoming;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.Serializable;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.jbi.component.ComponentContext;
import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.InOptionalOut;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerException;

import org.apache.axiom.attachments.Attachments;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMNamespace;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;
import org.apache.axiom.soap.SOAPBody;
import org.apache.axiom.soap.SOAPEnvelope;
import org.apache.axiom.soap.SOAPFactory;
import org.apache.axiom.soap.SOAPHeader;
import org.apache.axis2.AxisFault;
import org.apache.axis2.client.Options;
import org.apache.axis2.context.MessageContext;
import org.apache.axis2.engine.AxisEngine;
import org.apache.axis2.receivers.AbstractMessageReceiver;
import org.apache.axis2.util.Utils;
import org.objectweb.petals.binding.soap.listener.SUProperties;
import org.objectweb.petals.common.descriptor.Extensions;
import org.objectweb.petals.common.util.StringHelper;
import org.objectweb.petals.common.util.XMLUtil;
import org.objectweb.petals.component.common.util.PetalsExtensionsUtil;
import org.objectweb.petals.component.common.util.SourceHelper;
import org.w3c.dom.Document;

/**
 * The message receiver used by the binding component. The SOAP message is
 * forwarded to the JBI endpoint.
 *
 * @version $Rev: 250 $ $Date: 2006-04-21 14:20:57 +0200 (ven, 21 avr 2006) $
 * @since Petals 1.0
 * @author alouis,wjoseph,chamerling - EBM Websourcing
 *
 */
public class PetalsReceiver extends AbstractMessageReceiver {

    protected static final String ERROR_CREATE_RESPONSE_FROM_DOM = "Error creating a string from a DOM service description";

    protected static final String ERROR_CREATE_STRING_FROM_NM = "Error creating a string from normalized message";

    protected static final String ERROR_GET_WSDL = "Error while retrieving the service WSDL";

    protected static final String ERROR_SERVICE_INVOKATION = "An error occured in service invokation";

    protected static final String ERROR_SERVICE_NOT_FOUND = "Requested service and/or operation not found/available";

    protected static final String ERROR_WRONG_MESSAGE_STATUS = "JBI message has wrong status";

    protected DeliveryChannel channel;

    protected ComponentContext context;

    protected MessageExchangeFactory exchangeFactory;

    protected Logger logger;

    /**
     * Creates a new instance of PetalsReceiver
     *
     * @param context
     * @param channel
     * @param bindingSUM
     * @param log
     */
    public PetalsReceiver(ComponentContext context, DeliveryChannel channel,
            Logger log) {
        super();
        this.channel = channel;
        this.context = context;
        this.logger = log;
        this.exchangeFactory = channel.createExchangeFactory();
    }

    /*
     * (non-Javadoc)
     *
     * @see org.apache.axis2.engine.MessageReceiver#receive(org.apache.axis2.context.MessageContext)
     */
    public final void receive(final MessageContext msgContext) throws AxisFault {
        logger.log(Level.FINE, "Receiving a SOAP message from Axis2 engine");

        MessageContext outMsgContext = Utils
                .createOutMessageContext(msgContext);
        outMsgContext.getOperationContext().addMessageContext(outMsgContext);

        process(msgContext, outMsgContext);

        // send the response if required
        if (outMsgContext.getEnvelope() != null) {
            AxisEngine engine = new AxisEngine(msgContext.getOperationContext()
                    .getServiceContext().getConfigurationContext());

            engine.send(outMsgContext);
        }
    }

    /**
     * Create a String result from a DOM document
     *
     * @param document
     *            the DOM Document
     * @return a String representation of the DOM Document
     * @throws TransformerException
     */
    public String createStringFromDOMDocument(Document document)
            throws TransformerException {
        return XMLUtil.createStringFromDOMNode(document);
    }

    /**
     * Sends the SOAPEnvelope of the resquest message directly to a JBI service
     * and operation on the JBI container. The response message from the JBI
     * service invocation is transmitted as an outgoing message.
     *
     * @param inMessage
     *            MessageContext for incoming message
     * @param outMessage
     *            MessageContect fot outgoing message
     * @throws AxisFault
     *             error while processing request, send the JBI Exchange,
     *             processing the JBI response, or if the JBI response is a
     *             Fault
     */
    public void process(MessageContext inContext, MessageContext outContext)
            throws AxisFault {
        logger.log(Level.FINE, "Processing the incoming SOAP message");

        // retrieve Axis service from context
        PetalsAxisService axisService = null;
        if (inContext.getAxisService() instanceof PetalsAxisService) {
            axisService = (PetalsAxisService) inContext.getAxisService();
        } else {
            throw new AxisFault(
                    "Can not retrieve a valid PEtALS Axis service from context");
        }

        // Find the JBI endpoint
        ServiceEndpoint jbiEndpoint = findJBIEndpoint(axisService);
        if (jbiEndpoint == null) {
            throw new AxisFault("JBI endpoint not found");
        }

        // Retrieve message operation and service
        QName jbiOperation = retrieveSOAPOperation(inContext);
        if (jbiOperation == null) {
            throw new AxisFault("Operation not found in message context");
        }

        // remove the root element or not
        boolean remove = retrieveRemoveRootElement(axisService.getConsumes()
                .getExtensions());

        // get timeout
        long timeout = retrieveTimeOut(axisService.getConsumes().getExtensions());

        // create the message exchange
        MessageExchange exchange = null;
        try {
            exchange = createMessageExchange(inContext, jbiEndpoint,
                    jbiOperation, remove);
        } catch (AxisFault e) {
            String errorMsg = "Error while transforming SOAP request to JBI MessageExchange";
            logger.log(Level.SEVERE, errorMsg, e);
            throw new AxisFault(errorMsg, e);
        }

        // send the message exchange through JBI
        // As the SOAP/HTTP request may wait for a response,
        // use the sendSynchrone MEP
        boolean sent = false;
        try {
            sent = channel.sendSync(exchange, timeout);
        } catch (MessagingException e) {
            String errorMsg = "Error while sending message through JBI NMR.";
            logger.log(Level.SEVERE, errorMsg, e);
            throw new AxisFault(errorMsg, e);
        }

        if (!sent) {
            throw new AxisFault("The message has not been sent through the JBI NMR.");
        }

        // process the response received from JBI NMR. The outcontext is updated
        processJBIResponse(exchange, outContext, getSOAPFactory(inContext));
    }

    /**
     * Create the message exchange
     *
     * @param inContext
     * @param endpoint
     * @param operationName
     * @return
     * @throws AxisFault
     */
    protected MessageExchange createMessageExchange(MessageContext inContext,
            ServiceEndpoint endpoint, QName operationName, boolean removeRoot)
            throws AxisFault {
        InOptionalOut exchange = null;
        NormalizedMessage nm = null;

        try {

            // create an inoptional out message
            // TODO : get good MEP (feature request 306593)
            Source source = createSourceContent(inContext, removeRoot);

            // exchange = exchangeFactory.createExchange(mep);
            exchange = exchangeFactory.createInOptionalOutExchange();
            nm = exchange.createMessage();
            exchange.setInMessage(nm);
            nm.setContent(source);
            exchange.setEndpoint(endpoint);
            exchange.setOperation(operationName);

            // if (exchange instanceof InOut) {
            // InOut mex = (InOut) exchange;
            // mex.setInMessage(nm);
            // } else if (exchange instanceof InOptionalOut) {
            // InOptionalOut mex = (InOptionalOut) exchange;
            // mex.setInMessage(nm);
            // } else if (exchange instanceof InOut) {
            // InOut mex = (InOut) exchange;
            // mex.setInMessage(nm);
            // } else if (exchange instanceof RobustInOnly) {
            // RobustInOnly mex = (RobustInOnly) exchange;
            // mex.setInMessage(nm);
            // }

            // add SOAP attachements to normalized message
            copySOAPAttachmentsToJBI(inContext, nm);

            // Get the options from the SOAP message and put them into the JBI
            // message
            mapProperties(inContext, exchange);

        } catch (Exception e) {
            throw AxisFault.makeFault(e);
        }

        return exchange;
    }

    /**
     * Create a new Axis 2 SOAP body with the right name spaces and service name
     *
     * @param factory
     *            the soap factory
     * @param envelope
     *            a soap envelope
     * @param response
     *            the response
     * @return a new soap body
     */
    protected OMElement createSOAPBody(SOAPFactory factory,
            SOAPEnvelope envelope, String response) throws AxisFault {
        OMElement bodyContent = null;
        ByteArrayInputStream inputStream = new ByteArrayInputStream(response
                .getBytes());
        try {
            XMLStreamReader parser = XMLInputFactory.newInstance()
                    .createXMLStreamReader(inputStream);
            StAXOMBuilder builder = new StAXOMBuilder(factory, parser);
            bodyContent = (OMElement) builder.getDocumentElement();
        } catch (Exception e) {
            throw new AxisFault(
                    "Error parsing the response from JBI service to a SOAPBody",
                    e);
        }
        SOAPBody respBody = factory.createSOAPBody(envelope);
        respBody.addChild(bodyContent);
        return respBody;
    }

    /**
     * Create a new Axis 2 SOAP envelope with the right name spaces
     *
     * @param factory
     *            the soap factory; there are ex
     * @param namespaces
     *            name spaces for the soap envelope
     * @return a new SOAP envelope
     */
    protected SOAPEnvelope createSOAPEnvelope(SOAPFactory factory) {
        return factory.createSOAPEnvelope();
    }

    /**
     * Creates a SOAP response from a NormalizedMessage
     *
     * @param factory
     *            soap factory
     * @param nm2
     *            NormalizedMessage containing the response
     * @return a SOAPEnveloppe created from the nm NomalizedMessage content
     * @throws AxisFault
     */
    protected SOAPEnvelope createSOAPResponse(SOAPFactory factory,
            NormalizedMessage nm) throws AxisFault {

        SOAPEnvelope responseEnv = createSOAPEnvelope(factory);

        /*
         * Create and fill the Soap body with the content of the Normalized
         * message
         */
        String response = null;
        if (nm != null) {
            try {
                response = createStringContent(nm);
            } catch (Exception e) {
                logger.log(Level.SEVERE, ERROR_CREATE_STRING_FROM_NM);
                throw new AxisFault(ERROR_CREATE_STRING_FROM_NM, e);
            }
        } else {
            response = "<Response>Done</Response>";
        }
        createSOAPBody(factory, responseEnv, response);
        return responseEnv;
    }

    /**
     * Return the JBI ServiceEndpoint according to : <br>{
     * axisService.getTargetNamespace() } axisService.getName() <br>
     * axisService.getEndpoint()
     *
     * @param msgContext
     * @return the corresponding endpoint, null if no one corresponding
     */
    protected ServiceEndpoint findJBIEndpoint(PetalsAxisService axisService) {
        return context.getEndpoint(axisService.getConsumes().getServiceName(),
                axisService.getConsumes().getEndpointName());
    }

    /**
     * Return the operation according to the WSAAction SOAP attribute
     *
     * @param inContext
     *            SOAPrequest context
     * @return operation as QName, null if no WSAAction is specified
     */
    protected QName retrieveSOAPOperation(MessageContext inContext) {
        QName operationName = null;
        String operation = inContext.getWSAAction();
        if (!StringHelper.isNullOrEmpty(operation)) {
            operationName = QName.valueOf(operation);
        } else {
            operationName = retrieveOperationFromSOAPBody(inContext);
        }
        return operationName;
    }

    /**
     * Retrieve the remove root element extension value
     *
     * @param extensions
     * @return
     */
    protected boolean retrieveRemoveRootElement(Extensions extensions) {
        String result = PetalsExtensionsUtil.extractValueFromKeyValueExtension(
                extensions, SUProperties.REMOVE_ROOT, "false");
        return Boolean.parseBoolean(result);
    }

    /**
     *
     * @param extensions
     * @return
     */
    protected long retrieveTimeOut(Extensions extensions) {
        long result = 0L;
        String tmp = PetalsExtensionsUtil.extractValueFromKeyValueExtension(
                extensions, SUProperties.TIMEOUT, "0");
        try {
            result = Long.parseLong(tmp);
        } catch (NumberFormatException e) {
        }
        return result;
    }

    /**
     * Add the options properties from the Axis message context to the
     * properties of the normalized message.
     *
     * @param context
     * @param message
     */
    protected void mapProperties(MessageContext context,
            MessageExchange exchange) {
        Options options = context.getOptions();

        // get serializable properties, and add them to the message exchange
        // properties
        Map properties = options.getProperties();
        for (Object key : properties.keySet()) {
            Object value = properties.get(key);
            logger.fine("Adding SOAP property : " + key.toString() + " : "
                    + value.toString());
            if (value instanceof Serializable) {
                exchange.setProperty(key.toString(), properties.get(key));
            }
        }

        // get the SOAP header form envelope and add it as exchange property if
        // it exists
        try {
            SOAPEnvelope env = context.getEnvelope();
            SOAPHeader header = env.getHeader();
            if (header != null) {
                ByteArrayOutputStream os = new ByteArrayOutputStream();
                header.serialize(os);
                String s = os.toString();
                exchange.setProperty("SOAP.HEADER", s);
            }
        } catch (XMLStreamException e) {
            logger
                    .warning("Could not get SOAPHeader, it will not be added to message exchange");
        }
    }

    /**
     * Process the JBI response. <br>
     * According to the response status :
     * <li>
     * <ul>
     * do nothin for DONE,
     * </ul>
     * <ul>
     * throws an AxisFault for ERROR,
     * </ul>
     * <ul>
     * set outContext body to a FAULT,
     * </ul>
     * <ul>
     * set outContext body to the response.
     * </ul>
     * </li>
     *
     * @param exchange
     *            JBI response
     * @param outMessage
     *            set to FAULT or OUT JBI response
     * @param factory
     *            soap factory used to create body response
     * @throws AxisFault
     *             with ERROR JBI status, or problem while creating response
     */
    protected void processJBIResponse(MessageExchange exchange,
            MessageContext outMessage, SOAPFactory factory) throws AxisFault {
        if (exchange.getStatus().equals(ExchangeStatus.DONE)) {
            // exchange DONE, nothing to do, no soap response

        } else if (exchange.getStatus().equals(ExchangeStatus.ERROR)) {
            // exchange ERROR: throws an axis fault
            throw new AxisFault(exchange.getError());

        } else if (exchange.getStatus().equals(ExchangeStatus.ACTIVE)) {
            if (exchange.getFault() != null) {
                // exchange FAULT : return a fault message
                try {
                    outMessage.setEnvelope(createSOAPResponse(factory, exchange
                            .getFault()));
                } catch (AxisFault e) {
                    String errorMsg = "Error while creating SOAP Fault response.";
                    logger.log(Level.SEVERE, errorMsg, e);
                    throw new AxisFault(errorMsg, e);
                }

            } else {
                // Get the ouput message if the message is an instance of out
                try {
                    NormalizedMessage nm = null;
                    if (exchange instanceof InOptionalOut) {
                        InOptionalOut mex = (InOptionalOut) exchange;
                        nm = mex.getOutMessage();
                    } else if (exchange instanceof InOut) {
                        InOut mex = (InOut) exchange;
                        nm = mex.getOutMessage();
                    }

                    outMessage.setEnvelope(createSOAPResponse(factory, nm));

                    copyJBIAttachmentsToSOAP(nm, outMessage);
                } catch (AxisFault e) {
                    String errorMsg = "Error while creating SOAP Fault response.";
                    logger.log(Level.SEVERE, errorMsg, e);
                    throw new AxisFault(errorMsg, e);
                }
            }
            // close the messageExchange with the JBI NMR
            try {
                exchange.setStatus(ExchangeStatus.DONE);
                channel.send(exchange);
            } catch (MessagingException e) {
                String errorMsg = "Error while closing JBI MessageExchange.";
                logger.log(Level.SEVERE, errorMsg, e);
                throw new AxisFault(errorMsg, e);
            }

        } else {
            throw new AxisFault(ERROR_WRONG_MESSAGE_STATUS);
        }
    }

    /**
     * Add the normalized message attachments to the Axis message context.
     *
     * @param nm
     * @param messageContext
     *
     * @throws AxisFault
     */
    private void copyJBIAttachmentsToSOAP(NormalizedMessage nm,
            MessageContext messageContext) {
        if (nm != null && messageContext != null) {
            Set attachmentNames = nm.getAttachmentNames();
            for (Object id : attachmentNames) {
                String name = (String) id;
                messageContext.addAttachment(name, nm.getAttachment(name));
            }
        }
    }

    /**
     * Add the attachements from the incoming soap request to the Normalized
     * message
     *
     * @param messageContext
     * @param nm
     */
    private void copySOAPAttachmentsToJBI(MessageContext messageContext,
            NormalizedMessage nm) throws MessagingException {

        if (nm != null && messageContext != null) {
            Attachments attachments = messageContext.getAttachmentMap();
            if (attachments != null) {
                Set set = attachments.getContentIDSet();
                for (Object object : set) {
                    String id = (String) object;
                    DataHandler dh = attachments.getDataHandler(id);
                    nm.addAttachment(id, dh);
                }
            }
        }
    }

    /**
     * Create a new Source object containing the axis message Soap envelope
     *
     * @param message
     *            the axis message
     * @param removeRoot
     *            if root element of the SOAP body have to be removed
     * @return the source content
     * @throws Exception
     */
    private Source createSourceContent(MessageContext message,
            boolean removeRoot) throws Exception {
        String source = getBodyElement(message, removeRoot);
        return SourceHelper.createSource(source);
    }

    /**
     * Get the operation from the message context. Assume that the operation is
     * always the root element of the SOAP body.
     *
     * @param message
     * @return null if the operation has not been found
     */
    protected QName retrieveOperationFromSOAPBody(MessageContext message) {
        QName result = null;
        SOAPEnvelope env = message.getEnvelope();
        SOAPBody body = env.getBody();
        OMElement operation = body.getFirstElement();
        if (operation != null) {
            result = operation.getQName();
        }
        return result;
    }

    /**
     * Create a new String containing the normalized message content
     *
     * @param message
     *            the normalized message
     * @return the String content
     * @throws Exception
     */
    private String createStringContent(NormalizedMessage message)
            throws Exception {
        return SourceHelper.createString(message.getContent());
    }

    /**
     * Get the body from the SOAP message context. Remove the SOAP operation
     * (root element) if requested.
     *
     * @param message
     * @param removeSOAPAction
     *            true if the SOAP action has to be removed from the SOAP body.
     *            It means that the first element of the body is the SOAP
     *            action.
     * @return
     * @throws Exception
     */
    @SuppressWarnings("unchecked")
    private String getBodyElement(MessageContext message,
            boolean removeSOAPAction) throws Exception {
        String result = new String();

        SOAPEnvelope env = message.getEnvelope();
        SOAPBody body = env.getBody();
        OMNamespace namespace = env.getNamespace();
        Iterator<OMNamespace> envNS = env.getAllDeclaredNamespaces();
        Iterator<OMNamespace> bodyNS = body.getAllDeclaredNamespaces();

        OMElement rootElement = body.getFirstElement();
        if (removeSOAPAction) {
            // the first element is supposed tp be the operation, skip it from
            // the JBI payload
            rootElement = rootElement.getFirstElement();
        }
        if (rootElement != null) {
            rootElement.declareNamespace(namespace);
            while (envNS.hasNext()) {
                rootElement.declareNamespace(envNS.next());
            }
            while (bodyNS.hasNext()) {
                rootElement.declareNamespace(bodyNS.next());
            }
            ByteArrayOutputStream os = new ByteArrayOutputStream();
            rootElement.serialize(os);
            result = os.toString();
        }
        return result;
    }
}
