/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.binding.soap.listener.incoming;

import java.net.SocketException;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.apache.axis2.AxisFault;
import org.apache.axis2.Constants;
import org.apache.axis2.addressing.EndpointReference;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.description.Parameter;
import org.apache.axis2.description.TransportInDescription;
import org.apache.axis2.engine.ListenerManager;
import org.apache.axis2.transport.TransportListener;
import org.apache.axis2.transport.http.server.HttpUtils;
import org.mortbay.jetty.Server;
import org.mortbay.jetty.handler.ContextHandlerCollection;
import org.mortbay.jetty.servlet.Context;
import org.mortbay.jetty.servlet.ServletHolder;

/**
 * Create a Jetty server that will be the axis2 transportlistener. This Jetty
 * server is used to replace the SimpleHTTPServer provided by axis that have
 * quite bad performances.
 *
 * @author Christophe Hamerling - EBM WebSourcing
 *
 */
public class SoapServer implements TransportListener {

    /**
     * The logger
     */
    protected Logger logger;

    /**
     * The axis2 configuration context (built from axis2.xml config file)
     */
    protected ConfigurationContext configurationContext;

    /**
     * The component properties
     */
    protected SoapServerConfig config;

    /**
     * The Jetty server
     */
    private Server server;

    /**
     * The host address
     */
    private String hostAddress = null;

    /**
     * Creates a new instance of SoapServer
     *
     * @param configurationContext
     * @param port
     */
    public SoapServer(ConfigurationContext configContext,
        SoapServerConfig config) throws AxisFault {
        this.config = config;
        this.configurationContext = configContext;
        this.server = new Server(this.config.getPort());

        // create context handlers
        ContextHandlerCollection contexts = new ContextHandlerCollection();
        server.setHandler(contexts);

        // create the axis context
        Context axis2 = new Context(contexts,
            "/" + config.getServicesContext(), Context.SESSIONS);

        // create servlet holders
        ServletHolder axisServlet = new ServletHolder(new SoapServlet(
            configurationContext));
        axisServlet.setName("AxisServlet");
        axisServlet.setInitOrder(1);

        // add our own simple service list servlet
        ServletHolder listServlet = new ServletHolder(new ListServicesServlet(
            configurationContext, config.isProvidesList()));
        listServlet.setName("ServicesListServlet");
        listServlet.setInitOrder(1);
        axis2.addServlet(listServlet, "/" + config.getServicesMapping()
            + "/listServices");

        // do servlet mappings
        axis2.addServlet(axisServlet, "/" + config.getServicesMapping() + "/*");
        axis2.addServlet(axisServlet, "/servlet/AxisServlet");

        configureListenerManager();
    }

    /**
     *
     * @return
     */
    public ConfigurationContext getConfigurationContext() {
        return this.configurationContext;
    }

    /*
     * (non-Javadoc)
     *
     * @see org.apache.axis2.transport.TransportListener#getEPRsForService(java.lang.String,
     *      java.lang.String)
     */
    public EndpointReference getEPRForService(String serviceName, String ip)
        throws AxisFault {
        return getEPRsForService(serviceName, ip)[0];
    }

    /*
     * (non-Javadoc)
     *
     * @see org.apache.axis2.transport.TransportListener#getEPRsForService(java.lang.String,
     *      java.lang.String)
     */
    public EndpointReference[] getEPRsForService(String serviceName, String ip)
        throws AxisFault {
        // if host address is present
        if (hostAddress != null) {
            if (server != null) {
                return new EndpointReference[] {new EndpointReference(
                    hostAddress + "/"
                        + configurationContext.getServiceContextPath() + "/"
                        + serviceName)};
            } else {
                throw new AxisFault(
                    "Unable to generate EPR for the transport : http");
            }
        }
        // if the host address is not present
        String localAddress;
        if (ip != null) {
            localAddress = ip;
        } else {
            try {
                localAddress = HttpUtils.getIpAddress();
                if (localAddress == null) {
                    localAddress = "127.0.0.1";
                }
            } catch (SocketException e) {
                throw AxisFault.makeFault(e);
            }
        }
        if (server != null) {
            return new EndpointReference[] {new EndpointReference("http://"
                + localAddress + ":" + config.getPort() + "/"
                + configurationContext.getServiceContextPath() + "/"
                + serviceName)};
        } else {
            throw new AxisFault(
                "Unable to generate EPR for the transport : http");
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.apache.axis2.transport.TransportListener#init(org.apache.axis2.context.ConfigurationContext,
     *      org.apache.axis2.description.TransportInDescription)
     */
    public void init(ConfigurationContext axisConf,
        TransportInDescription transprtIn) throws AxisFault {
        try {
            this.configurationContext = axisConf;

            Parameter param = transprtIn.getParameter(PARAM_PORT);

            if (param != null) {
                this.config
                    .setPort(Integer.parseInt((String) param.getValue()));
            }
            param = transprtIn.getParameter(HOST_ADDRESS);
            if (param != null) {
                hostAddress = ((String) param.getValue()).trim();
            }
        } catch (Exception e) {
            throw AxisFault.makeFault(e);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.apache.axis2.transport.TransportListener#start()
     */
    public void start() throws AxisFault {
        try {
            server.start();
        } catch (Exception e) {
            logger.log(Level.SEVERE, "Can not start the Jetty server");
            throw new AxisFault("Can not start the Jetty server", e);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see org.apache.axis2.transport.TransportListener#stop()
     */
    public void stop() throws AxisFault {
        try {
            // stop jetty server
            server.stop();
        } catch (InterruptedException e) {
            throw AxisFault.makeFault(e);
        } catch (Exception e) {
            throw AxisFault.makeFault(e);
        }
    }

    /**
     * Configure the listener manager used by Axis
     *
     * @throws AxisFault
     */
    private void configureListenerManager() throws AxisFault {
        ListenerManager listenerManager = configurationContext
            .getListenerManager();
        TransportInDescription trsIn = new TransportInDescription(new QName(
            Constants.TRANSPORT_HTTP));
        trsIn.setReceiver(this);
        if (listenerManager == null) {
            listenerManager = new ListenerManager();
            listenerManager.init(configurationContext);
        }
        listenerManager.addListener(trsIn, true);
    }
}
